// Encapsulates SVG items, has tag, properties, content and children
function Item(tag, properties, content) {
	this._tag = tag;
	this.properties = properties;
	this._content = content;
	this._children = [];
	
	function setPreamble(pre) {
		// Stuff before the <svg tag
		this._preamble = pre;
	}
		
	function addChild(child) {
		if (child) {
			this._children.push(child);
			child._parent = this;			
		}
	};

	function getTag() {
		return this._tag;
	};
	
	// Gets the property (or state)
	function getProperty(name) {
		if (!properties) {
			return;
		}
		name += '=';
		var roleInd = this.properties.indexOf(name);
		if (roleInd == -1) {
			return;
		}
		
		var role = this.properties.substring(roleInd + name.length);
		var list = role.split('"');
		return list[1];		
	};
	
	function getRole() {
		return this.getProperty('role');
	};

	// Writes out the SVG, if unaltered, this should be a duplicate of the original
	function asString(sp) {
		sp = (sp) ? sp : 0;
		var pre = getSpaces(sp) + "<" + tag ;
		if (this._preamble) {
			pre = this._preamble + "\n" + pre;
		}
		if (this.properties) {
			pre = pre + " " + this.properties + " ";
		}
		if (this._content) {
			pre = pre + ">" + this._content + "</" + tag + ">\n";
		} else if (this._children.length == 0) {
			pre = pre + "/>\n";
		} else {
			pre = pre + ">\n";
			for (var i = 0; i < this._children.length; i++) {
				pre = pre + this._children[i].asString(sp + 1);
			}
			pre = pre + getSpaces(sp) + "</" + tag + ">\n";
		}
		
		return pre;		
	};
	
	function getSpaces(sp) {
		// For formatting SVG by level
		var str = "";
		for (var i = 0; i < sp; i++) {
			str = str + "  ";
		}
		return str;
	};
	
	function getChildren() {
		return this._children;
	}
	
	// Exposing functions
	this.addChild = addChild;
	this.asString = asString;
	this.getTag = getTag;
	this.setPreamble = setPreamble;
	this.getChildren = getChildren;
	this.getRole = getRole;
	this.getProperty = getProperty;
};

// Temporary object used to build the svg scene tree from a SVG string
function Branch(tag, index, data) {
	this.tag = tag;
	this.index = index;
	this.data = data;
	this.comment = tag.indexOf('!--') == 0;
	this.openTag = (this.comment) ? false :  data.cArray[index].indexOf("/>") == -1;
	this.end = index;
	this.children = (this.openTag) ? [] : null;
};

// Builds a Item tree for the svg string
function svgToTree(svg) {
	var info = {"svg": svg};
	var preamble = getPreamble(info);
	var str = info.svg.trim();
	var array = str.split("<");	
	var tagArray = [];
	var tagMap = mapTags(array, tagArray);
	var length = array.length;
	var data = { "index" : 1,
				 "cArray": array,
				 "tagArray": tagArray,
				 "tagMap" : tagMap
				};
	makeBranches(data);
	var root = convertBranchesToItems(data.branches);
	if (preamble) {
		root.setPreamble(preamble);
	}
	return root;
};

// Converts an array of Branch objects to Item objects 
function convertBranchesToItems(branches) {
	var length = branches.length;
	var root = null;
	for(var i = 0; i < length; i++) {
		// Only needs to find the first branch which is the root of the SVG
		if (branches[i]) {
			root = branches[i];
		}
	}
	
	var item = makeItem(root);
	mapStyles(item);
	return item;
};

// Makes an Item object from a Branch object
function makeItem(branch) {
	var str = branch.data.cArray[branch.index];

	if (branch.openTag) {
		var index = str.indexOf('>');
		var properties = str.substring(0, index);
		var content = str.substring(index + 1);
		var c = content.trim();
		if (c.length == 0) {
			content = null;
		}
		var item = new Item(branch.tag, properties, content);
		var numChildren = branch.children.length;
		for(var i = 0; i < numChildren; i++) {
			var child = makeItem(branch.children[i]);
			item.addChild(child);
		}
		return item;
	}
	
	if (branch.comment) {
		return;
	}
	var cIndex = str.indexOf('/>');
	
	var props = (cIndex > 0) ? str.substring(0, cIndex) : null;
	return new Item(branch.tag, props, null);	
};

function makeBranches(data) {
	var branches = [];
	var length = data.cArray.length;
	for(var i = 0; i < length; i++) {
		var tag = data.tagArray[i];
		if (tag && tag.charAt(0) != '/') {
			branches[i] = new Branch(tag, i, data);
		}
	}
	data.branches = branches;

	// Find the ends of the branches
	for(var i = length - 1; i > -1; i--) {
		if (branches[i] && branches[i].openTag) {
			findBranchEnd(data, branches, i);
		}
	}
	
	// Set the parent for the branches
	for(var i = length - 1; i > -1; i--) {
		if (branches[i]) {
			setParent(branches, i);
		}
	}
	
}

function setParent(brs, index) {
	var bStart = brs[index].index;
	
	for(var i = index - 1; i > -1; i--) {
		if (brs[i] && brs[i].end > bStart) {
			brs[i].children.unshift(brs[index]);
			brs[index] = null;
			return;
		}
	}
};

function findBranchEnd(data, branches, i) {	
	var closers = data.tagMap['/'+ branches[i].tag];
	if (!closers) {
		var tag = branches[i].tag;
		if (tag == '![CDATA[') {
			data.cArray[i - 1] = data.cArray[i - 1] + '<' + tag + data.cArray[i] + '>';
			delete branches[i];
			return;
		}
	}
	if (closers) {
		var clen = closers.length;
		var branchIndex = branches[i].index;
		for(var j = 0; j < clen; j++) {
			if (closers[j] > branchIndex) {
				// This is the close tag for the branch
				branches[i].end = closers[j];
				closers.splice(j, 1);
				return;
			}
		}		
	}
};

function findMatchingCloseTag(tag, data) {
	// If no children, the next tag should be the close tag
	var nextTag = data.tagArray[data.index + 1];
	if (nextTag && nextTag.substring(1) == tag) {
		return data.index + 1;
	}
	
	// Must have children 
	var closeTagArray = data.tagMap["/"+tag];
	if (closeTagArray && closeTagArray.length == 1) {
		// Only one close tag
		return closeTagArray[0];
	}
	var tagArray = data.tagMap[tag];
	var hello = "hello";  // good for a breakpoint
	
};

// Gets stuff that may before <svg, like xml and doctype declarations
function getPreamble(info) {
	var str = info.svg;
	var end = str.indexOf("?>");
	var docTypeIndex = str.indexOf('<!DOCTYPE');

	if (end != -1 || docTypeIndex != -1) {
		var preamble = null;
		if (docTypeIndex > end) {
			var string = str.substring(docTypeIndex);
			var index = string.indexOf('>');
			preamble = str.substring(0, docTypeIndex + index + 1);
			info.svg = str.substring(docTypeIndex + index + 1);
		} else {
			preamble = str.substring(0, end + 2);
			info.svg = str.substring(end + 2);			
		}
		return preamble;
	}
	return;
};

// Makes a map (key = tag, value = cArray index) for working out tag nesting
// Populates the tagArray and truncates the tag from the array strings
function mapTags(array, tagArray) {
	// array is the svg string split by '>'
	var map = {};
	var length = array.length;
	for(var i = 0; i < length; i++) {
		var str = array[i].trim();
		if (str.length == 0) {
			array[i] = str; // catches any empty splits (first split is always empty)
		} else {
			// Testing for tag and property split "\\s"
			var space = str.search("\\s");
			var rightAngle = str.indexOf(">");
			
			// Extract the tag, logic based on where the tag ends in the string
			if (space > 0 && space < rightAngle) {
				// Have whitespace tag is on left of space
				extractTag(space, str, i, map, tagArray, array);
			} else {
				if (rightAngle > 0) {
					// Test for closed tag /> vs  an open tag >
					var closedTag = str.indexOf("/>");
					if (closedTag > 0) {
						extractTag(closedTag, str, i, map, tagArray, array);
					} else {
						extractTag(rightAngle, str, i, map, tagArray, array);
					}
				}
			}
		}
	}
	return map;
};


function extractTag(tagEnd, str, i, map, tagArray, array) {
	var tag = str.substring(0, tagEnd);
	var tagIndices = map[tag];
	if (tagIndices == null) {
		tagIndices = [];
		map[tag] = tagIndices;
	}
	tagIndices.push(i);
	tagArray[i] = tag;

	array[i] = str.substring(tagEnd + 1); 
};

function mapStyles(item) {
	var tagMap = {};
	mapItemTags(item, tagMap);
	var styles = tagMap['style'];
	if (!styles) {
		return;
	}
			
	extractStyles(styles, tagMap, item);
};

function extractStyles(styleList, tagMap, item) {
	var classMap = {};
	var idMap = {};
	for(var i = 0; i < styleList.length; i++) {
		var style = styleList[i];
		var content = style._content;
		if (!content) {
			continue;
		}
		var styles = content.split('}');
		for (var j = 0; j < styles.length; j++) {
			var selectorAndProps = styles[j].split('{');
			if (selectorAndProps.length < 2) {
				continue;  // not a selector and properties
			}
			var selector = selectorAndProps[0].trim();
			if (selector.indexOf('.') == 0) {
				// class selector
				selector = selector.substring(1);
				classMap[selector] = style;
			} else if (selector.indexOf('#') == 0) {
				selector = selector.substring(1);
				idMap[selector] = style;				
			} else {
				// tag selector
				var items = tagMap[selector];
				if (items) {
					for(var k = 0; k < items.length; k++) {
						items[k].style = {"type": style };
					}
				}
				
			}
		}
	}
	
	if (Object.keys(classMap).length > 0) {
		addStyleSelectors(classMap, item, 'class','classid' );
	}
	if (Object.keys(idMap).length > 0) {
		addStyleSelectors(idMap, item, 'id','idid' );
	}
};

function addStyleSelectors(map, item, attribute, key) {
	var index = item.properties.indexOf(attribute);
	if (index != -1) {
		var rest = item.properties.substring(index + attribute.length);
		splits = rest.split('"');
		var style = map[splits[1]];
		if (item.style) {
			item.style[key] = style;
		} else {
			item.style = {};
			item.style[key] = style;
		}		
	}
	
	var children = item.getChildren();
	for(var i = 0; i < children.length; i++) {
		addStyleSelectors(map, children[i], attribute, key);
	}
}

function mapItemTags(item, tagMap) {
	var tag = item.getTag();
	var tagList = tagMap[tag];
	if (!tagList) {
		tagList = [];
		tagMap[tag] = tagList;
	}
	tagList.push(item);
	
	var children = item.getChildren();
	var length = children.length;
	for(var i = 0; i < length; i++) {
		mapItemTags(children[i], tagMap);
	}		
};


