// Object for building a tree of the SVG
// padX = padding left and right, padY = padding top and bottom, rect width, rect height - width and height of rectangle
// vgap - vertical gap between rectangles in the tree, hgap - horizontal gap between rectangles in the tree
function NodeBuilder(padX, padY, rectWidth, rectHeight, vgap, hgap) {
	this.level = 0;
	this.x = padX;
	this.y = padY;
	this.rectWidth = rectWidth;
	this.rectHeight = rectHeight;
	this.maxX = padX;
	this.maxY = padY;
	this.padX = padX;
	this.padY = padY;
	this.vgap = vgap;
	this.hgap = hgap;
	this.rectClose = '</rect>';
	this.group = '<g><defs>';
	var style = '<style type="text/css"> rect { fill-opacity: 1; }';
     style += ' .link { stroke:blue; stroke-width:2px; } .label { fill:black; font-family:Arial,SanSerif; font-size:1.2em; } ';
	style +=  ' .tlink { stroke:#CCCCFF; stroke-width:2px; }  .clink { stroke:#FFCCFF; stroke-width:2px; }  .ilink { stroke:#FF9933; stroke-width:2px; } ';
	style += ' .role { font-weight:bold; font-family:Arial,SanSerif; font-size:1.2em; }';
	style += ' .aria { font-style:italic; font-family:Arial,SanSerif; font-size:1.2em;}';
	style += '</style></defs>';
	this.group += style;
	this.groupClose = '</g>';	
	
	
	// Internal function to create a rectangle for drawing
	function createRect(loc, width, fill, title) {
		if (loc.x + this.rectWidth > this.maxX) {
			this.maxX = loc.x + width;
		}

		if (loc.y + this.rectHeight > this.maxY) {
			this.maxY = loc.y+ this.rectHeight;
		}
		var rect = '<rect x=' + loc.x + ' y=' + loc.y + ' height=' + this.rectHeight + ' width=' + width + ' fill=' + fill + '>';
		if (title) {
			rect += ' <title>' + title + '</title> ';
		}
		return rect + this.rectClose;
	};
	// Internal function to create the label for drawing
	function makeText(text, loc) {
		var xt = loc.x + 5;
		var yt = loc.y + this.rectHeight - 5;
		var txt = '<text x=' + xt + ' y=' + yt + ' class="label">' + text + '</text>';
		return txt;
	};
	
	// Internal function to layout the SVG as a tree, item is the root node
	function layout(item) {
		// lays out the tree
		var levels = {};
		this.addLocations(item, 0, levels); // Add loc objects, with x
		
		// find the most populated level (targetLevel)
		var maxLevel = levels[0];
		var targetLevel = 0;		
		var numLevels = Object.keys(levels).length;
		for (var level = 0; level < numLevels; level++) {
			var list = levels[level];
			if (list.length >= maxLevel.length) {
				maxLevel = list;
				targetLevel = level;
			} 
		}
		
		// Layout the tallest level in y
		var delta = this.vgap + this.rectHeight;
		for(var i = 0; i < maxLevel.length; i++) {
			maxLevel[i].loc.y = this.padY + (i * delta);
		}
		
		// Iterate level toward root, populating the y location from children
		var leafMap = {};
		var parentLevel = targetLevel - 1;
		while (parentLevel > -1) {
			var leaveList = [];
			var members = levels[parentLevel];
			for(var i = 0; i < members.length; i++) {
				var childless = propLocFromChildren(members[i]);
				if (childless) {
					leaveList.push(childless);
				}
			}
			leafMap[parentLevel] = leaveList;
			parentLevel--;
		}
		
		// Iterate away from the root, propagating y values
		var branchLevel = targetLevel + 1;
		for (var i = branchLevel; i < numLevels; i++) {
			var levelMaxY = 0;  // Max y used (so far) on this level
			var level = levels[i];
			var length = level.length;
			for(var j = 0; j < length; j++) {
				var item = level[j];
				if (!item.loc.y) {
					levelMaxY = this.setYFromParent(item, levelMaxY);
				}
			}
		}
		
		// Set y in skipped leaves, and propagate y's to a leaf's branch if needed
		var leafLevel = targetLevel - 1;
		while (leafLevel > 0) {
			var leaves = leafMap[leafLevel];
			var fullList = levels[leafLevel];
			this.calcYFromNeighbors(leaves, fullList);
			leafLevel--;
			if (leafLevel > -1) {
				// Before calculating from neighbors, calc from children
				leaves = leafMap[leafLevel];
				for(var i = 0; i < leaves.length; i++) {
					propLocFromChildren(leaves[i]);
				}
				
			}
		}
				
		if (!levels[0][0].loc.y) {
			// svg tag needs a y coord
			var tl = levels[targetLevel];
			var last = tl[tl.length - 1];
			levels[0][0].loc.y = last.loc.y / 2;
		}
	};
	
	// Internal function to calculate the y coordinate for a node
	function calcYFromNeighbors(leafList, list) {	
		var start = 0;
		var delta = this.vgap + this.rectHeight;
		for (var i = 0; i < leafList.length; i++) {
			var leaf = leafList[i];
			if (leaf.loc.y) {
				continue;  // Already done
			}
			for (var j = start; j < list.length; j++) {
				if (leaf == list[j]) {
					leaf.loc.y = (j == 0) ? this.padY : list[j - 1].loc.y + delta;	
					start = j+1;
					break;
				}
				
			}
		}
		
		resolveCollisions(list, delta);
	}
	
	// Internal function, shifts nodes if they overlap
	function resolveCollisions(list, delta) {
		var length = list.length;
		for(var i = 1; i < length; i++) {
			var y = list[i - 1].loc.y;
			if (list[i].loc.y < y + delta) {
				list[i].loc.y = y + delta;
			}
		}
	}
	
	// Internal function, sets Y value from parent node
	function setYFromParent(item, levelMaxY) {
		var parent = item._parent;
		var siblings = parent.getChildren();
		var y = parent.loc.y;
		if (siblings.length == 1) {
			if (y > levelMaxY) {
				item.loc.y = y;
				return y;
			} else {
				y = levelMaxY + this.vgap + this.rectHeight;
				item.loc.y = y;
				return y;
			}
		}
		
		var height = (this.vgap + this.rectHeight) * siblings.length;
		y -= height/2;
		if (y <= levelMaxY) {
			y = levelMaxY + this.vgap + this.rectHeight;
		}
		for (var i = 0; i < siblings.length; i++) {
			siblings[i].loc.y = y;
			y += this.vgap + this.rectHeight;
		}
		
		return y;
		
	};
	
	// Internal function calculates y from a node's children
	function propLocFromChildren(item) {
		var children = item.getChildren();
		if (children.length == 0) {
			return item;
		}
		
		var minY = children[0].loc.y;
		var maxY = children[children.length - 1].loc.y;
		if (minY && maxY) {
			item.loc.y = (minY + maxY) / 2;
		} else {
			return item;
		}	
		
	};
	
	// Internal function - adds the location object the item
	function addLocations(item, level, counter) {
		var x = this.padX + (level * (this.rectWidth + this.hgap));
		item.loc = { "level": level, "x": x, "counter": counter};
		var list = counter[level];
		if (!list) {
			list = [];
			counter[level] = list;
		}
		list = list.push(item);
		
		var children = item.getChildren();
		for(var i = 0; i < children.length; i++) {
			this.addLocations(children[i], level + 1, counter);
		}
	};
	
	// External function - draws the SVG node as a tree, item should be an svgTree.js Item object
	function drawTree(item) {
		this.layout(item);
		var links = this.createLinks(item);
		var rects = this.createRectsAndText(item);
		var str =  '<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" height=' + this.getHeight() + ' width=' + this.getWidth() + '>\n';
		str += '  ' + this.group + '\n    ' + links + rects + this.groupClose + '\n</svg>';
		return str;
	};
	
	
	function getRoleSymbol(role) {
		if (!role) {
			return;
		}
		
		var symbol = null;
		switch(role) {
		case 'data':
			symbol ='d';
			break;
		case 'axis':
			symbol = 'a';
			break;
		case 'majoritem':
			symbol = 'ma';
			break;
		case 'none':
			symbol = "x";
			break;
		case 'label':
			symbol = 'L';
			break;
		case 'item':
			symbol = 'i';
			break;
		case 'title':
			symbol = 't';
			break;
		case 'legend':
			symbol = 'le';
			break;
		case 'symbol':
			symbol = 's';
			break;
		case 'style':
			symbol = 'st';
			break;
		case 'grid':
			symbol = 'gr';
			break;
		case 'guide':
			symbol = 'gu';
			break;
		case 'group':
			symbol = 'g';
			break;
		case 'connector':
			symbol = 'c';
			break;
		case 'event':
			symbol = 'e';
			break;
		case 'note':
			symbol = 'n';
			break;
		case 'background':
			symbol = 'b';
			break;
		case 'view':
			symbol = 'v';
			break;

		default:
			symbol = '?';
		}
		return {role: role, symbol: symbol};
	}
	
	
	// Gets the rectangle's fill color 
	function getFill(tag) {
		var fill = '"#FFCC00"';
		switch(tag) {
		case 'g' :
			fill = '"#ADFF5C"';
			break;
		case 'svg' :
			fill = '"#99FF33"';
			break;
		case 'style':
			fill = '"#FFCCFF"';
			break;
		case 'title':
			// fall through
		case 'desc':
			fill = '"#E0EBCC"';
			break;
		case 'defs':
			fill = '"#CCCCFF"';
			break;
	
		case 'metadata': 
			fill = '"#99CCFF"';
			break;
			
		case 'stop':
			fill = '"#EBD6FF"';
			break;
		case 'filter':
			fill = '"#FF99FF"';
			break;
		case 'feGaussianBlur':
			fill = '"#FFE0FF"';
			break;
		case 'use':
			fill = '"#B2F0FF"';
			break;			
			
		default:
			if (tag.indexOf('rdf') == 0 || tag.indexOf('cc') == 0 ||  tag.indexOf('dc') == 0) {
				fill = '"#D6EBFF"';
			} else if (tag.indexOf('Gradient') > -1) {
				fill = '"#CC99FF"';
			}
		}		
		return fill;
	}
	
	// Creates the drawn rectangle and label
	function createRectsAndText(item) {
		var str = '';
		if (item.loc.y) {
			var tag = item.getTag();
			var fill = getFill(tag);
			var title = (tag == 'title' || tag=='desc') ? item._content : item.properties;
			str += this.createRect(item.loc, this.rectWidth, fill, title);
			str += this.addStyleBox(item);
			str += this.makeText(tag, item.loc);
			str += this.addTextContent(item);
		}
		str += "\n";
		var role = getRoleSymbol(item.getRole());
		if (role) {
			var sym = '<text x=' + (item.loc.x + 10) + ' y=' + item.loc.y + ' class="role">' + role.symbol;
			sym += '<title>role = ' + role.role +  '</title></text>';
			str += sym;
		}
		
		var labeledBy = item.getProperty('aria-labeledby');
		if (labeledBy) {
			var sym = '<text x=' + (item.loc.x + 45) + ' y=' + item.loc.y + ' class="aria">' + 'lb';
			sym += '<title>aria-labeledby = ' + labeledBy +  '</title></text>';
			str += sym;
			
		}
		
		var children = item.getChildren();
		for(var i = 0; i < children.length; i++) {
			str += this.createRectsAndText(children[i]);
		}
		return str;
	};
		
	// Creates the drawn links between nodes (rectangles)
	function createLinks(item) {
		var children = item.getChildren();
		var str = '';
		if (item.loc.y) {
			var parentAnchorX = item.loc.x + this.rectWidth - 2;
			var parentAnchorY = item.loc.y + (this.rectHeight / 2);
			var childAnchorY = parentAnchorY;

			var base = '<line x1=' + parentAnchorX + ' y1=' + parentAnchorY + ' x2=';
			str += this.makeStyleLinks(item, parentAnchorY);

			for (var i = 0; i < children.length; i++) {
				var child = children[i];
				if (!child.loc.y) {
					continue;
				}
				var y = child.loc.y + this.rectHeight / 2;
				str += base + (child.loc.x + 2) + ' y2=' + y + ' class="link" />\n';					
			}
		}

		for (var i = 0; i < children.length; i++) {
			var child = children[i];
			var cLink = this.createLinks(child);
			if (cLink) {
				str += cLink;
			}
		}
		
		return str;
	};
	
	function makeStyleLinks(item, itemAnchorY) {
		if (!item.style) {
			return '';
		}

		var str = '';
		if (item.style.classid) {
			var loc = item.style.classid.loc;
			str = this.addStyleLink(loc, item, itemAnchorY, "clink");			
		}		
		if (item.style.type) {
			var loc = item.style.type.loc;
			str += this.addStyleLink(loc, item, itemAnchorY, "tlink");
		}
		if (item.style.idid) {
			var loc = item.style.idid.loc;
			str = this.addStyleLink(loc, item, itemAnchorY, "ilink");			
			
		}
		
		return str;
	}
	
	function addStyleLink(loc, item, itemAnchorY, classId) {
		if (!loc.y) {
			return '';
		}

		var x1 = null;
		var x2 = null;

		if (loc.x < item.loc.x) {
			x1 = item.loc.x - 4;
			x2 = loc.x + this.rectWidth + 4;
		} else if (loc.x > item.loc.x) {
			x1 = item.loc.x + this.rectWidth + 4;
			x2 = loc.x - 4;
		} else {
			x1 = loc.x + rectWidth + 4;
			x2 = x1;
		}
		var y2 = loc.y + this.rectHeight / 2;
		str = '<line x1=' + x1 + ' y1=' + itemAnchorY + ' x2=' + x2 + ' y2='
				+ y2;
		return str + ' class="' + classId + '" />\n';
	}
	
	function addStyleBox(item) {
		if ((!item.properties || item.properties.length < 1) && !item.style) {
			return '';
		} 
		
		var str = '';
		var dx = item.loc.x + this.rectWidth - 10;
		if (item.style) {
			if (item.style.classid) {
				var loc = { x: dx, y: item.loc.y, level: item.loc.level};
				var fill = '"#FFCCFF"';
				str += this.createRect(loc, 10, fill, 'style from class ');
				dx -= 10;							
			}
			if (item.style.type) {
				var loc = { x: dx, y: item.loc.y, level: item.loc.level};
				var fill = '"#CCCCFF"';
				str += this.createRect(loc, 10, fill, 'style from tag ');
				dx -= 10;							
			}
			if (item.style.id) {
				var loc = { x: dx, y: item.loc.y, level: item.loc.level};
				var fill = '"#FF9933"';
				str += this.createRect(loc, 10, fill, 'style from id ');
				dx -= 10;							
			}

		}
		
		var style = item.properties.indexOf('style');
		if (style > -1 ) {
			var loc = { x: dx, y: item.loc.y, level: item.loc.level};
			var fill = '"#8080FF"';
			str += this.createRect(loc, 10, fill, 'style in properties');
		}
		return str;
	};
	
	function addTextContent(item) {
		var tag = item.getTag();
		var str = '';
		if (item._content && (tag == 'text' || tag == 'tspan')) {
			var loc = {x: item.loc.x + this.rectWidth + 5, y: item.loc.y, level: item.loc.level};
			str = this.makeText(item._content, loc);
			var estimatedWidth = item._content.length * 8 + 5; // 5 from makeText
			if (this.maxX < estimatedWidth + loc.x) {
				this.maxX = estimatedWidth + loc.x;
			}
		}
		return str;
	}
	
	
	// Functions for setting the viewport
	function getWidth() {
		return this.maxX + this.padX;
	};
	
	function  getHeight() {
		return this.maxY + this.padY;
	};
	
	this.getWidth = getWidth;
	this.getHeight = getHeight;
	this.createRect = createRect;
	this.makeText = makeText;
	this.createLinks = createLinks;
	this.addLocations = addLocations;
	this.layout = layout;
	this.drawTree = drawTree;
	this.createRectsAndText = createRectsAndText;
	this.setYFromParent = setYFromParent;
	this.calcYFromNeighbors = calcYFromNeighbors;
	this.addStyleBox = addStyleBox;
	this.addTextContent = addTextContent;
	this.makeStyleLinks = makeStyleLinks;
	this.addStyleLink = addStyleLink;
};




