//
// Function treeNavigation() is a class constructor for a treev widget. The widget binds to an
// unordered list. The top-level <ul> must have role='tree'. All list items must have role='treeitem'.
//
// Tree groups must be embedded lists within the listitem that heads the group. the top <ul> of a group
// must have role='group'. aria-expanded is used to indicate whether a group is expanded or collapsed. This
// property must be set on the listitem the encapsulates the group.
//
// parent nodes must be given the class tree-parent.
//
// @param (treeID string) treeID is the html id of the top-level <ul> of the list to bind the widget to
//
// @return N/A
//
function treeNavigation(treeID) {

  // Find the navigable items
  var id = '#' + treeID;
  this.$svg = $(id).find('svg');
  var $svgWithHideRole = $(id).find('svg').not("[role='none']").not("[role='background']");
  this.svgVisible = (this.$svg.length == $svgWithHideRole.length);
  this.$activeItem = null; // holds the jQuery object for the active item

  this.keys = {
            tab:      9,
            enter:    13,
            space:    32,
            pageup:   33,
            pagedown: 34,
            end:      35,
            home:     36,
            left:     37,
            up:       38,
            right:    39,
            down:     40,
            asterisk: 106
   };


  // initialize the treeNavigation
  this.init();

  // bind event handlers
  this.bindHandlers();

} // end treeNavigation() constructor

//
// Function init() is a member function to initialize the treeNavigation widget. 
//
// @return N/A
//
treeNavigation.prototype.init = function() {
	// Find navigable objects and pass through groups 
	
	  // Objects with a first child of <title> or <desc> should be navigable, unless role of none or background
	  var $hasTitles = this.$svg.find(':first-child').filter('title').parent().not("[role='none']").not("[role='background']");
	  var $hasDesc = this.$svg.find(':first-child').filter('desc').parent().not("[role='none']").not("[role='background']");
	  
	   // Objects with roles should be navigable, unless role is none or background
	  var $withRoles = this.$svg.find('[role]').not("[role='none']").not("[role='background']");
	  if (this.svgVisible) {
		  $withRoles = $withRoles.add(this.$svg);
	  }

	  // Objects with aria-(label, labeledby, describedby) should be navigable
	  var $labeledBy = this.$svg.find('[aria-labeledby]');
	  var $describedBy = this.$svg.find('[aria-describedby]');
	  var $labels = this.$svg.find('[aria-label]');
	  
	  // everything so far
	  var $all = $withRoles.add($describedBy).add($labels).add($labeledBy).add($hasTitles).add($hasDesc);
	  // If the SVG does not have a none or background role, include the children of the SVG
	  
	  this.$groups = $all.filter('g');
	  
	  // pick up descendants of things with a role, that don't have a role
	  var $descendants = $withRoles.children('*').not('[role]').not('defs').not('metadata');
	  var $children = $descendants;
	  while($children.length > 0) {
		  $children = $children.children('*').not('[role]');
		  $descendants = $descendants.add($children);
	  }
	  
	  // remove groups, titles and desc, from descendants
	  // if a group should be navigated or exposed, it will be in $all
	  this.$passThroughGroups = $descendants.filter('g');
	  $descendants = $descendants.not('g').not('title').not('desc');  
	  
	  this.$navigableItems = $all.add($descendants).add(this.$svg);

} // end init()


//
// Function bindHandlers() is a member function to bind event handlers to the navigable items
//
// return N/A
//
treeNavigation.prototype.bindHandlers = function() {

  var thisObj = this;
  var $par = this.$svg.parent();
  
  // bind a click handler on each navigable item
  this.$navigableItems.click(function(e) {
    return thisObj.handleClick($(this), e);
  });

  // bind a keydown handler to the SVG object
  $par.keydown(function(e) {
    return thisObj.handleKeyDown(e);
  });

  // bind a keypress handler to the SVG object
  $par.keypress(function(e) {
    return thisObj.handleKeyPress(e);
  });

  // bind a focus handler to the SVG object
 
  $par.focus(function(e) {
    return thisObj.handleFocus(e);
  });
  $par.attr('tabindex', '0');
//  $par.attr('role', "application");

  // bind a blur handler to the SVG object
  this.$svg.blur(function(e) {
    return thisObj.handleBlur(e);
  });

   // bind a document click handler
   $(document).click(function(e) {

         if (thisObj.$activeItem != null) {      
        	thisObj.$activeItem.attr('tabindex', '-1');
            thisObj.$activeItem = null;
         }
         thisObj.$svg.attr('tabindex', '0');

         return true;
   });

} // end bindHandlers()


treeNavigation.prototype.isNavigable = function($item) {
	return (this.$navigableItems.index($item) > -1);
}

treeNavigation.prototype.isPassThru = function($item) {
	return (this.$passThroughGroups.index($item) > -1);
}

treeNavigation.prototype.isSelectable = function($item) {
	return (this.$navigableItems.index($item) > -1 || this.$passThroughGroups.index($item) > -1);
}

//
// Function handleKeyDown() is a member function to process keydown events for the treeNavigation items
//
// @param (e object) e is the associated event object
//
// @return (boolean) returns false if consuming event; true if not
//
treeNavigation.prototype.handleKeyDown = function(e) {

	if ((e.altKey || e.ctrlKey) || (e.shiftKey && e.keyCode != this.keys.tab)) {
		// do nothing
		return true;
	}

	switch (e.keyCode) {
//	case this.keys.tab: {
//		// set activeItem to null
//		this.$activeItem = null;
//		return true;
//	}
	case this.keys.home: { // jump to first item in the branch level
		this.setActiveItem(this.getFirstSibling());
		e.stopPropagation();
		return false;
	}
	case this.keys.end: { // jump to last visible item
		this.setActiveItem(this.getLastSibling());
		e.stopPropagation();
		return false;
	}
	case this.keys.enter:
	case this.keys.space: {
		e.stopPropagation();
		return false;
	}
	case this.keys.left: {
		// Move up the tree
		this.setActiveItem(this.toParent());
		e.stopPropagation();
		return false;
	}
	case this.keys.right: {
		// Move down a child
		if (this.$activeItem != this.$svg) {
			// debug code, allowing a breakpoint
			this.$temp = this.$activeItem;
		}

		this.setActiveItem(this.toChild(this.$activeItem));
		e.stopPropagation();
		return false;
	}
	case this.keys.up: {
		// Move to earlier sibling
		this.setActiveItem(this.previousSibling());
		e.stopPropagation();
		return false;
	}
	case this.keys.down: {
		// Move to later sibling
		this.setActiveItem(this.nextSibling());
		e.stopPropagation();
		return false;
	}
	case this.keys.asterisk: {

		e.stopPropagation();
		return false;
	}
	}

	return true;

} // end handleKeyDown


// TODO make a getSiblings function that takes into account pass through groups 
treeNavigation.prototype.getSiblings = function() {
	if (this.$activeItem == null || this.$activeItem == this.$svg) {
		return this.$svg;
	}
	
	
	// TODO needs work, try with airlines axes labels... almost :)
	var $ancestor = this.$activeItem.parent();
	while(this.isPassThru($ancestor)) {
		$ancestor = $ancestor.parent();
	}
	
	var $siblings = $ancestor.children();	
	var $passThruSibs = $ancestor.children('g').not('[role]');
	$siblings = $siblings.not($passThruSibs);
	
	while($passThruSibs.length > 0) {
		var $descendents = $passThruSibs.children().not('g');
		$siblings = $siblings.add($descendents);
		$groupSibs = $passThruSibs.children('g');
		$passThruSibs = $groupSibs.not('[role]');
		$gSibs = $groupSibs.not($passThruSibs);
		$siblings = $siblings.add($gSibs);
	}
	return $siblings;
}

treeNavigation.prototype.getLastSibling = function() {
	// Called when the END key used
	if (this.$activeItem == null || this.$activeItem == this.$svg) {
		return this.$svg;
	}
	
	var $siblings = this.getSiblings();

	// Find the last navigable sibling 
	var length = $siblings.length;
	for(var i = length - 1; i > -1; i--) {
		var $sib = $($siblings.get(i));
		if (this.isSelectable($sib)) {
			return $sib;
		}
	}
	
	return this.$activeItem;
}

treeNavigation.prototype.getFirstSibling = function() {
	// Called when the HOME key used
	if (this.$activeItem == null || this.$activeItem == this.$svg) {
		return this.$svg;
	}
		
	var $siblings = this.getSiblings();
	var length = $siblings.length;
	// Find the first navigable sibling 
	for(var i = 0; i < length; i++) {
		var $sib = $($siblings.get(i));
		if (this.isSelectable($sib)) {
			return $sib;
		}
	}
	
	return this.$activeItem;
}

treeNavigation.prototype.nextSibling = function() {
	// Called when the down arrow key used
	if (this.$activeItem == null || this.$activeItem == this.$svg) {
		return this.$svg;
	}
		
	var $siblings = this.getSiblings();

	// Find the next navigable sibling that is not the $activeItem
	var index = $siblings.index(this.$activeItem);
	var length = $siblings.length;
	for(var i = index + 1; i < length; i++) {
		var $sib = $($siblings.get(i));
		if (this.isSelectable($sib)) {
			return $sib;
		}
	}
	
	return this.$activeItem;	
}

treeNavigation.prototype.previousSibling = function() {
	// Called when the Up arrow key used
	if (this.$activeItem == null || this.$activeItem == this.$svg) {
		return this.$svg;
	}
	
	var $siblings = this.getSiblings();

	// Find the next navigable sibling that is not the $activeItem
	var index = $siblings.index(this.$activeItem);
	for(var i = index - 1; i > -1; i--) {
		var $sib = $($siblings.get(i));
		if (this.isSelectable($sib)) {
			return $sib;
		}
	}
	
	return this.$activeItem;	
}

treeNavigation.prototype.toParent = function() {
	// Called when the Up arrow key used
	if (this.$activeItem == null || this.$activeItem == this.$svg) {
		return this.$svg;
	}
	
	var $parent = this.$activeItem.parent();
	while (this.isPassThru($parent)) {
		$parent = $parent.parent();
	}

	return $parent;
}

treeNavigation.prototype.toChild = function($item) {
	if (this.$activeItem == null) {
		return this.$svg;
	}
	
	var $children = $item.children();
	var length = $children.length;
	for(var i = 0; i < length; i++) {
		var $child = $($children.get(i));
		if (this.isNavigable($child)) {
			return $child;
		} else if (this.isPassThru($child)) {
			return this.toChild($child);
		}
	}
	
	return $item;
}


//
// Function handleKeyPress() is a member function to process keypress events for the treeNavigation items
// This function is needed for browsers, such as Opera, that perform window manipulation on kepress events
// rather than keydown. The function simply consumes the event.
//
// @param (e object) e is the associated event object
//
// @return (boolean) returns false if consuming event; true if not
//
treeNavigation.prototype.handleKeyPress = function(e) {

  if (e.altKey || e.ctrlKey || e.shiftKey) {
    // do nothing
    return true;
  }

  switch (e.keyCode) {
    case this.keys.tab: {
      return true;
    }
    case this.keys.enter:
    case this.keys.home:
    case this.keys.end:
    case this.keys.left:
    case this.keys.right:
    case this.keys.up:
    case this.keys.down: {
      e.stopPropagation();
      return false;
    }
      default : {
//         var chr = String.fromCharCode(e.which);
//         var bMatch = false;
//         var itemNdx = this.$navigableItems.index($item);
//         var itemCnt = this.$navigableItems.length;
//         var curNdx = itemNdx + 1;
//
//         // check if the active item was the last one on the list
//         if (curNdx == itemCnt) {
//            curNdx = 0;
//         }
//
//         // Iterate through the menu items (starting from the current item and wrapping) until a match is found
//         // or the loop returns to the current menu item
//         while (curNdx != itemNdx)  {
//
//            var $curItem = this.$navigableItems.eq(curNdx);
//            var titleChr = $curItem.text().charAt(0);
//            
//            if ($curItem.is('.tree-parent')) {
//               titleChr = $curItem.find('span').text().charAt(0);
//            }
//
//            if (titleChr.toLowerCase() == chr) {
//               bMatch = true;
//               break;
//            }
//
//            curNdx = curNdx+1;
//
//            if (curNdx == itemCnt) {
//               // reached the end of the list, start again at the beginning
//               curNdx = 0;
//            }
//         }
//
//         if (bMatch == true) {
//            this.$activeItem = this.$navigableItems.eq(curNdx);
//            this.$activeItem.focus();
//         }
//
         e.stopPropagation();
         return false;
      }
  }

  return true;

} // end handleKeyPress


//
// Function handleClick() is a member function to process click events.
//
// @param ($id object) $id is the jQuery id of the parent item firing event
//
// @param (e object) e is the associated event object
//
// @return (boolean) returns false if consuming event; true if not
//
treeNavigation.prototype.handleClick = function($id, e) {

  if (e.altKey || e.ctrlKey || e.shiftKey || !this.isNavigable($id)) {
    // do nothing
    return true;
  }
  
  
//  this.$activeItem = $id;
  e.stopPropagation();
  return false;

} // end handleClick

treeNavigation.prototype.setActiveItem = function($item) {
	  // update the active item and give it focus
	this.$activeItem = $item;
	this.$navigableItems.attr('tabindex', '-1');
	// This is a bad thing to do, but it shows navigation in Firefox and Chrome
	this.$navigableItems.removeAttr("class");
	// removeClass is what we should be doing, but it does not affect the outerHTML 
	//  which needs to happen for SVG to pick up the CSS changes
//	this.$navigableItems.removeClass("activeElem");
	this.$svg.attr('tabindex', '0');
	if (this.$activeItem) {
		this.$activeItem.attr('tabindex', '0');
		// This is a bad thing to do, but it shows navigation in Firefox and Chrome
		this.$activeItem.attr('class', 'activeElem');
		// addClass is what we should be doing, but it does not affect the outerHTML
		// which needs to happen for SVG to pick up the CSS changes
//		this.$activeItem.addClass("activeElem");
		var elem = this.$activeItem.get(0);		
		this.$activeItem.focus();	
		console.log(elem);
	} else {
		console.log("active item null");
	}
}

//
// Function handleFocus() is a member function to process focus events.
//
// @param ($item object) $item is the jQuery id of the parent item firing event
//
// @param (e object) e is the associated event object
//
// @return (boolean) returns true
//
treeNavigation.prototype.handleFocus = function(e) {
  this.setActiveItem(this.$svg);
  return true;

} // end handleFocus

//
// Function handleBlur() is a member function to process blur events.
//
// @param ($id object) $id is the jQuery id of the parent item firing event
//
// @param (e object) e is the associated event object
//
// @return (boolean) returns true
//
treeNavigation.prototype.handleBlur = function($id, e) {
 // this.$activeItem = null;
	
	this.$navigableItems.attr('tabindex', '-1');
	this.$navigableItems.removeAttr("class");
	// removeClass is what we should be doing, but it does not affect the outerHTML 
	//  which needs to happen for SVG to pick up the CSS changes
	// We should be using removeClass
	this.$navigableItems.removeClass('activeElem');
	this.$svg.attr('tabindex', '0');
	return true;

} // end handleBlur